@@ -2,9 +2,12 @@
 from __future__ import absolute_import
 from .croniter import (
     croniter,
+    datetime_to_timestamp,
     croniter_range,
+    CroniterBadTypeRangeError,  # noqa
     CroniterBadDateError,  # noqa
     CroniterBadCronError,  # noqa
-    CroniterNotAlphaError  # noqa
+    CroniterNotAlphaError, # noqa
+    CroniterUnsupportedSyntaxError, #noqa
 )  # noqa
 croniter.__name__  # make flake8 happy
